#!/usr/bin/python

import gzip
import operator
import os
import shutil
import subprocess
import sys
import tempfile
import time
import urllib2

from debian_bundle import deb822, changelog

from launchpadlib.credentials import Credentials
from launchpadlib.launchpad import Launchpad, EDGE_SERVICE_ROOT

from bzrlib import branch, bzrdir, errors, tag, testament, trace, transport, urlutils
from bzrlib.plugin import load_plugins
from bzrlib.trace import mutter, log_exception_quietly, note

load_plugins()

from bzrlib.plugins.builddeb import import_dsc
from bzrlib.plugins.builddeb.util import find_changelog
from bzrlib.plugins.builddeb.errors import MissingChangelogError

base_dir = "/srv/package-import.canonical.com/new/"
if not os.path.exists(base_dir):
    base_dir = "."
packages_dir = os.path.join(base_dir, "new-packages")
updates_dir = os.path.join(base_dir, "updates")
lists_dir = os.path.join(base_dir, "lists")
logs_dir = os.path.join(base_dir, "logs")
revids_dir = os.path.join(base_dir, "revids")
lp_creds_file = os.path.join(base_dir, "lp_creds.txt")
lp_cache_dir = os.path.join(base_dir, "lp_cache")

if len(sys.argv) > 2:
    trace.set_verbosity_level(1)

def pool_base(name):
    if name.startswith("lib"):
        return name[:4]
    return name[0]

distros = ("debian", "ubuntu",)

distro_list_base = {"ubuntu": "http://archive.ubuntu.com/ubuntu/dists/",
    "debian": "http://ftp.debian.org/dists/",
}

distro_releases = {"ubuntu": ["warty", "hoary", "breezy", "dapper", "edgy", "feisty", "gutsy", "hardy", "intrepid", "jaunty", "karmic"],
    "debian": ["woody", "sarge", "etch", "lenny", "squeeze", "sid", "experimental"],
}

# development releases first for stacking, no etch or earlier as they aren't on lp
lp_distro_releases = {"ubuntu": ["karmic", "warty", "hoary", "breezy", "dapper", "edgy", "feisty", "gutsy", "hardy", "intrepid", "jaunty"],
    "debian": ["squeeze", "lenny", "sid", "experimental"],
}

distro_pockets = {"ubuntu": ["release", "security", "proposed", "updates", "backports"],
    "debian": ["release",],
}

distro_components = {"ubuntu": ["main", "restricted", "universe", "multiverse"],
    "debian": ["main", "contrib", "non-free"],
}

distro_pool = {"ubuntu": "http://archive.ubuntu.com/ubuntu/",
    "debian": "http://ftp.debian.org/",
}

launchpad_base_url = "https://launchpad.net/"
ubuntu_base_url = launchpad_base_url + "ubuntu/"
urllib_launchpad_base_url = "https+urllib://launchpad.net/"
urllib_debian_base_url = "http+urllib://ftp.uk.debian.org/debian/"
urllib_debian_archive_base_url = "http+urllib://archive.debian.org/debian/"


def debian_base_url(release):
    if release in ["hamm", "slink", "potato", "woody", "sarge"]:
        return urllib_debian_archive_base_url
    return urllib_debian_base_url


f = open(lp_creds_file, "rb")
try:
    creds = Credentials("package-import")
    creds.load(f)
finally:
    f.close()
lp = Launchpad(creds, service_root=EDGE_SERVICE_ROOT, cache=lp_cache_dir)
ubuntu = lp.distributions['ubuntu']
u_archive = ubuntu.main_archive


def get_debian_versions(package):
    proc = subprocess.Popen(["/usr/bin/madison-lite", "--mirror", lists_dir,
            package], stdout=subprocess.PIPE)
    ret = proc.wait()
    assert ret == 0, "Error running madison-lite"
    versions = []
    for line in proc.stdout:
        # package | version | release/component | arch
        ret_package = line[:line.index("|")].strip()
        assert ret_package == package
        line = line[line.index("|")+1:].strip()
        version = line[:line.index("|")].strip()
        line = line[line.index("|")+1:].strip()
        release_component = line[:line.index("|")].strip()
        if "/" in release_component:
            release = release_component[:release_component.index("/")]
            component = release_component[release_component.index("/")+1:]
        else:
            release = release_component
            component = "main"
        versions.append((changelog.Version(version), "debian", release, component, "release"))
    return versions


def get_versions(package):
    versions = []
    versions = get_debian_versions(package)
    publications = u_archive.getPublishedSources(source_name=package, exact_match=True)
    for publication in publications:
        assert publication.source_package_name == package
        release = publication.distro_series.name.lower()
        pocket = publication.pocket.lower()
        version = changelog.Version(publication.source_package_version)
        if (version, "ubuntu", release, "", pocket) not in versions:
            versions.append((version, "ubuntu", release, "", pocket))
    def compare_versions(a, b):
        if a[1] == "ubuntu" and a[1] == b[1]:
            if a[4] == "backports" and b[4] != "backports":
                return -1
            if a[4] != "backports" and b[4] == "backports":
                return 1
        if a[0] < b[0]:
            return 1
        if a[0] > b[0]:
            return -1
        if a[1] != b[1]:
            if a[1] == "debian":
                return 1
            return -1
        if a[2] != b[2]:
            a_release = distro_releases[a[1]].index(a[2])
            b_release = distro_releases[b[1]].index(b[2])
            if a_release < b_release:
                return 1
            return -1
        if a[4] != b[4]:
            a_pocket = distro_pockets[a[1]].index(a[4])
            b_pocket = distro_pockets[b[1]].index(b[4])
            if a_pocket < b_pocket:
                return 1
            return -1
        return 0
    versions.sort(cmp=compare_versions)
    return versions


def branch_location(distro, package, release, suite):
    distro, suite = translate_lp_suite(distro, suite)
    return "bzr+ssh://james-w@bazaar.launchpad.net/~ubuntu-branches/%s/%s/%s/%s" % (distro, release, package, suite)


def create_branch(path):
    to_transport = transport.get_transport(path)
    format = bzrdir.format_registry.make_bzrdir("2a")
    to_transport.ensure_base()
    new_branch = bzrdir.BzrDir.create_branch_convenience(
            to_transport.base, format=format,
            possible_transports=[to_transport])
    return new_branch


def create_updates_branch(target_branch_path, updates_branch_path):
    try:
        br_from = branch.Branch.open(target_branch_path)
    except errors.NotBranchError:
        updates_branch = create_branch(updates_branch_path)
        return updates_branch
    to_transport = transport.get_transport(updates_branch_path)
    # Branch the branch in to the updates area
    br_from.lock_read()
    try:
        rev_id = br_from.last_revision()
        to_transport.mkdir('.')
        dir = br_from.bzrdir.sprout(to_transport.base,
            rev_id, possible_transports=[to_transport])
        updates_branch = dir.open_branch()
        tag._merge_tags_if_possible(br_from, updates_branch)
    finally:
        br_from.unlock()
    return updates_branch


def other_distro(distro):
    assert distro in ("debian", "ubuntu")
    if distro == "debian":
        return "ubuntu"
    else:
        return "debian"


def grab_file(location, target_dir):
    mutter("fetching %s" % location)
    location_base = urlutils.dirname(location)
    location_file = urlutils.basename(location)
    local_path = os.path.join(target_dir, location_file)
    def get_file(transport):
        return transport.get(location_file)
    def redirected(transport, e, redirection_notice):
        target = urlutils.unescape(e.target).encode('utf-8')
        return transport._redirected_to(e.source, target)
    location_f = transport.do_catching_redirections(get_file,
                        transport.get_transport(location_base),
                        redirected)
    local_f = open(local_path, "wb")
    try:
        shutil.copyfileobj(location_f, local_f)
    finally:
        local_f.close()


def dget(dsc_location, target_dir):
    grab_file(dsc_location, target_dir)
    local_dsc_path = os.path.join(target_dir,
        urlutils.basename(dsc_location))
    dsc_f = open(local_dsc_path)
    try:
        dsc = deb822.Dsc(dsc_f)
        files = dsc['files']
        for file_info in files:
            name = file_info['name']
            # TODO: md5 check
            grab_file(urlutils.join(urlutils.dirname(dsc_location), name),
                target_dir)
    finally:
        dsc_f.close()
    return local_dsc_path


def push_branch(br_from, to_loc, possible_transports=None):
    created = False
    to_transport = transport.get_transport(to_loc,
                possible_transports=possible_transports)
    try:
        dir_to = bzrdir.BzrDir.open_from_transport(to_transport)
    except errors.NotBranchError:
        dir_to = None
    if dir_to is None:
        br_to = br_from.create_clone_on_transport(to_transport)
        created = True
    else:
        dir_to.push_branch(br_from)
    return created


def make_db_set(name, distro, release, pocket, updates_branch):
    db_set = import_dsc.DistributionBranchSet()
    for d in distros:
        for known_release in distro_releases[d]:
           for known_pocket in distro_pockets[d]:
               suite = make_suite(known_release, known_pocket)
               db = None
               if d == distro and known_release == release and known_pocket == pocket:
                   db = import_dsc.DistributionBranch(updates_branch, None)
                   db.tree = updates_branch.bzrdir.open_workingtree()
                   the_db = db
               else:
                   other_branch_path = branch_location(d, name, release, suite)
                   try:
                       other_branch = branch.Branch.open(other_branch_path)
                   except errors.NotBranchError:
                       other_branch = None
                   if other_branch is not None:
                       db = import_dsc.DistributionBranch(other_branch, other_branch)
               if db is not None:
                   db_set.add_branch(db)
    return the_db


def make_full_db_set(temp_dir, name, distro, release, pocket, updates_branch, possible_transports=None):
    db_set = import_dsc.DistributionBranchSet()
    for d in distros:
        for known_release in distro_releases[d]:
           for known_pocket in distro_pockets[d]:
               suite = make_suite(known_release, known_pocket)
               db = None
               if d == distro and known_release == release and known_pocket == pocket:
                   db = import_dsc.DistributionBranch(updates_branch, None)
                   db.tree = updates_branch.bzrdir.open_workingtree()
                   the_db = db
               else:
                   other_branch_path = os.path.join(temp_dir, suite)
                   try:
                       other_branch = branch.Branch.open(other_branch_path, possible_transports=possible_transports)
                   except errors.NotBranchError:
                       other_branch = None
                   if other_branch is not None:
                       db = import_dsc.DistributionBranch(other_branch, other_branch)
               if db is not None:
                   db_set.add_branch(db)
    return the_db


def make_suite(release, pocket):
    if pocket == "release":
        suite = release
    elif pocket in ("security", "proposed", "updates", "backports"):
        suite = "%s-%s" % (release, pocket)
    else:
        assert False, "Unkown pocket: %s" % pocket
    return suite


def import_package(name, version, distro, release, pocket, location):
    suite = make_suite(release, pocket)
    target_branch_path = branch_location(distro, name, release, suite)
    if not os.path.exists(updates_dir):
        os.makedirs(updates_dir)
    temp_dir = tempfile.mkdtemp(prefix="builddeb-", suffix="-" + name,
            dir=updates_dir)
    updates_branch_path = os.path.join(temp_dir, "%s-%s-%s" % (name, distro, suite))
    try:
        format = bzrdir.format_registry.make_bzrdir('2a')
        to_transport = transport.get_transport(temp_dir)
        to_transport.ensure_base()
        newdir = format.initialize_on_transport(to_transport)
        repo = newdir.create_repository(shared=True)
        repo.set_make_working_trees(True)
        updates_branch = create_updates_branch(target_branch_path, updates_branch_path)
        update_db = make_db_set(name, distro, release, pocket, updates_branch)
        if update_db.has_version(version):
            mutter("Already imported %s" % str(version))
            return
        local_dsc_path = dget(location, temp_dir)
        update_db.import_package(local_dsc_path)
        mark_revid(name, version, update_db.revid_of_version(version), suite,
                   update_db.branch)
        push_branch(updates_branch, target_branch_path)
    finally:
        shutil.rmtree(temp_dir)


def full_import_package(temp_dir, name, version, distro, release, pocket, location, possible_transports=None):
    suite = make_suite(release, pocket)
    updates_branch_path = os.path.join(temp_dir, suite)
    try:
        updates_branch = branch.Branch.open(updates_branch_path, possible_transports=possible_transports)
    except errors.NotBranchError:
        updates_branch = create_branch(updates_branch_path)
    update_db = make_full_db_set(temp_dir, name, distro, release, pocket, updates_branch, possible_transports=possible_transports)
    upstream_dir = tempfile.mkdtemp(dir=temp_dir)
    upstream_tip = None
    tags = updates_branch.tags.get_tag_dict().items()
    timestamps = {}
    filtered_tags = []
    for tag, revid in tags:
        if not tag.startswith("upstream-"):
            continue
        try:
            revobj = updates_branch.repository.get_revision(revid)
        except errors.NoSuchRevision:
            continue
        else:
            timestamp = revobj.timestamp
        timestamps[revid] = timestamp
        filtered_tags.append(tag, revid)
    filtered_tags.sort(key=lambda x: timestamps[x[1]])
    if len(filtered_tags) > 0:
        upstream_tip = filtered_tags[-1]
    if upstream_tip is not None:
        update_db._extract_upstream_tree(upstream_tip, upstream_dir)
    else:
        update_db._create_empty_upstream_tree(upstream_dir)
    try:
        dl_dir = tempfile.mkdtemp()
        try:
            local_dsc_path = dget(location, temp_dir)
            update_db.import_package(local_dsc_path,
                    use_time_from_changelog=True)
        finally:
            shutil.rmtree(dl_dir)
    finally:
        shutil.rmtree(upstream_dir)
    mark_revid(name, version, update_db.revid_of_version(version), suite,
               update_db.branch)


def mark_revid(package, version, revid, suite, branch):
    revid_map = load_revids(package, suite)
    branch.lock_read()
    try:
        tment = testament.StrictTestament3.from_revision(branch.repository, revid)
        sha = tment.as_sha1()
    finally:
        branch.unlock()
    assert str(version) not in revid_map, \
          "Asked to mark %s %s, but it is already marked" \
                % (package, version)
    revid_map[str(version)] = (revid, sha)
    revid_file = revid_filename(package, suite)
    f = open(revid_file, "wb")
    try:
        for (v, (r, s)) in sorted(revid_map.items()):
            f.write("%s %s %s\n" % (v, r, s))
    finally:
        f.close()


def revid_filename(package, suite):
    revid_dir = os.path.join(revids_dir, package)
    if not os.path.exists(revid_dir):
        os.makedirs(revid_dir)
    revid_file = os.path.join(revid_dir, suite)
    return revid_file


def load_revids(package, suite):
    revid_file = revid_filename(package, suite)
    revid_map = {}
    if os.path.exists(revid_file):
        f = open(revid_file)
        try:
            for line in f:
                v, r, s = line.split()
                revid_map[v] = (r, s)
        finally:
            f.close()
    return revid_map


def check_revid(package, version, revid, suite, branch):
    revid_map = load_revids(package, suite)
    tment = testament.StrictTestament3.from_revision(branch.repository, revid)
    sha = tment.as_sha1()
    assert str(version) in revid_map, \
          "%s %s %s is not marked" % (package, str(version), suite)
    assert (revid, sha) == revid_map[str(version)], \
         "%s != %s for %s %s, something has changed" \
             % (str((revid, sha)), str(revid_map[str(version)]), package,
                          str(version))


def find_unimported_versions(all_versions, name):
    # Deal with collisions.
    unimported_versions = []
    imported_suites = set()
    for (version, release, pocket) in reversed(all_versions):
        suite = make_suite(release, pocket)
        target_branch_path = branch_location(distro, name, release, suite)
        try:
            target_branch = branch.Branch.open(target_branch_path)
            has_version = True
        except errors.NotBranchError:
            has_version = False
        if has_version:
            db = import_dsc.DistributionBranch(target_branch, target_branch)
            has_version = db.has_version(version)
        if not has_version:
            unimported_versions.insert(0, (version, release, pocket))
        else:
            #imported_suites.add(suite)
            revid = db.revid_of_version(version)
            check_revid(name, version, revid, suite, db.branch)
    return unimported_versions


def translate_lp_suite(distro, suite):
    if distro == "debian":
        if suite == "oldstable":
            suite = "etch"
        elif suite == "stable":
            suite = "lenny"
        elif suite == "testing":
            suite = "squeeze"
        elif suite == "unstable":
            suite = "sid"
    return distro, suite


def lp_branch_url(distro, package, release, suite):
    distro, suite = translate_lp_suite(distro, suite)
    return 'https://api.edge.launchpad.net/beta/~ubuntu-branches/%s/%s/%s/%s' % (distro, release, package, suite)


def full_import(package, *extra):
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)
    log_name = os.path.join(logs_dir, package)
    f = open(log_name, "ab")
    log_token = trace.push_log_file(f)
    try:
        if not os.path.exists(updates_dir):
            os.makedirs(updates_dir)
        temp_dir = tempfile.mkdtemp(prefix="builddeb-", suffix="-" + package,
                    dir=updates_dir)
        try:
            format = bzrdir.format_registry.make_bzrdir('2a')
            to_transport = transport.get_transport(temp_dir)
            to_transport.ensure_base()
            newdir = format.initialize_on_transport(to_transport)
            repo = newdir.create_repository(shared=True)
            repo.set_make_working_trees(True)
            versions = get_versions(package)
            assert len(versions) > 0
            versions.reverse()
            mutter(str(versions))
            possible_transports = []
            for (version, distro, release, component, pocket) in versions:
                if version.debian_version is not None:
                    version_str = str(version.upstream_version) + "-" + str(version.debian_version)
                else:
                    version_str = str(version.upstream_version)
                if distro == "ubuntu":
                    files_url = urllib_launchpad_base_url + "ubuntu/%s/+source/%s/%s/+files/" % (release, package, str(version))
                else:
                    start = package[0]
                    if package.startswith("lib"):
                        start = package[:4]
                    files_url = debian_base_url(release) + "pool/%s/%s/%s/" % (component, start, package)
                package_url = files_url + "%s_%s.dsc" % (package, version_str)
                mutter("import %s to %s %s-%s (%s)" % (package_url, distro, release, pocket, component))
                full_import_package(temp_dir, package, version, distro, release, pocket, package_url, possible_transports=possible_transports)
            for d in distros:
                for known_release in lp_distro_releases[d]:
                    for known_pocket in distro_pockets[d]:
                        suite = make_suite(known_release, known_pocket)
                        try:
                            source_branch = branch.Branch.open(os.path.join(temp_dir, suite))
                        except errors.NotBranchError:
                            continue
                        to_loc = branch_location(d, package, known_release, suite)
                        created = push_branch(source_branch, to_loc, possible_transports=possible_transports)
                        branch_url = lp_branch_url(d, package, known_release, suite)
                        lp_dist = lp.distributions[d]
                        lp_series = lp_dist.getSeries(name_or_version=known_release)
                        lp_sp = lp_series.getSourcePackage(name=package)
                        lp_branch = lp.load(branch_url)
                        lp_sp.setBranch(branch=lp_branch, pocket=known_pocket[0].upper() + known_pocket[1:])
                        if known_release == "squeeze" or known_release == "karmic":
                            time.sleep(300)
            print "Imported %s" % str(versions)
            return 0
        finally:
            shutil.rmtree(temp_dir)
            pass
    except:
        log_exception_quietly()
        raise
    finally:
        trace.pop_log_file(log_token)
        f.close()


def main(package, *extra):
    if not os.path.exists(logs_dir):
        os.makedirs(logs_dir)
    log_name = os.path.join(logs_dir, package)
    f = open(log_name, "ab")
    log_token = trace.push_log_file(f)
    try:
        versions = get_versions(package)
        assert len(versions) > 0
        versions.reverse()
        mutter(str(versions))
        versions = find_unimported_versions(versions, package)
        mutter("These versions are new: %s" % str(versions))
        if len(versions) == 0:
            print "Nothing new"
            return 0
        for (version, distro, release, component, pocket) in versions:
            if version.debian_version is not None:
                version_str = str(version.upstream_version) + "-" + str(version.debian_version)
            else:
                version_str = str(version.upstream_version)
            if distro == "ubuntu":
                files_url = urllib_launchpad_base_url + "ubuntu/%s/+source/%s/%s/+files/" % (release, package, str(version))
            else:
                start = package[0]
                if package.startswith("lib"):
                    start = package[:4]
                files_url = debian_base_url(release) + "pool/%s/%s/%s/" % (component, start, package)
            package_url = files_url + "%s_%s.dsc" % (package, version_str)
            mutter("import %s to %s %s-%s (%s)" % (package_url, distro, release, pocket, component))
            import_package(package, version, distro, release, pocket, package_url)
        print "Imported %s" % str(versions)
        return 0
    except:
        log_exception_quietly()
        raise
    finally:
        #try:
        #    status_file = file('/proc/%s/status' % os.getpid(), 'rb')
        #except IOError:
        #    pass
        #else:
        #    status = status_file.read()
        #    status_file.close()
        #    note("Process status after command:")
        #    for line in status.splitlines():
        #        note(line)
        trace.pop_log_file(log_token)
        f.close()

if len(sys.argv) < 2:
    print "Must pass the name of one package to import"
    sys.exit(1)

#sys.exit(main(*sys.argv[1:]))
sys.exit(full_import(*sys.argv[1:]))

